# GPIO Factory helper
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
from __future__ import annotations
import os
import re
import asyncio
import pathlib
import logging
import periphery
from ..utils import KERNEL_VERSION

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Optional
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..eventloop import EventLoop

GpioEventCallback = Callable[[float, float, int], Optional[Awaitable[None]]]

GPIO_PATTERN = r"""
    (?P<bias>[~^])?
    (?P<inverted>!)?
    (?:(?P<chip_id>gpiochip[0-9]+)/)?
    (?P<pin_name>gpio(?P<pin_id>[0-9]+))
"""

BIAS_FLAG_TO_DESC: Dict[str, str] = {
    "^": "pull_up",
    "~": "pull_down",
    "*": "disable" if KERNEL_VERSION >= (5, 5) else "default"
}

class GpioFactory:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.reserved_gpios: Dict[str, GpioBase] = {}

    def setup_gpio_out(self, pin_name: str, initial_value: int = 0) -> GpioOutputPin:
        initial_value = int(not not initial_value)
        pparams = self._parse_pin(pin_name, initial_value)
        gpio = self._request_gpio(pparams)
        try:
            gpio_out = GpioOutputPin(gpio, pparams)
        except Exception:
            logging.exception("Error Instantiating GpioOutputPin")
            gpio.close()
            raise
        full_name = pparams["full_name"]
        self.reserved_gpios[full_name] = gpio_out
        return gpio_out

    def register_gpio_event(
        self, pin_name: str, callback: GpioEventCallback
    ) -> GpioEvent:
        pin_params = self._parse_pin(pin_name, req_type="event")
        gpio = self._request_gpio(pin_params)
        event_loop = self.server.get_event_loop()
        try:
            gpio_event = GpioEvent(event_loop, gpio, pin_params, callback)
        except Exception:
            logging.exception("Error Instantiating GpioEvent")
            gpio.close()
            raise
        full_name = pin_params["full_name"]
        self.reserved_gpios[full_name] = gpio_event
        return gpio_event

    def _request_gpio(self, pin_params: Dict[str, Any]) -> periphery.GPIO:
        full_name = pin_params["full_name"]
        if full_name in self.reserved_gpios:
            raise self.server.error(f"GPIO {full_name} already reserved")
        chip_path = pathlib.Path("/dev").joinpath(pin_params["chip_id"])
        if not chip_path.exists():
            raise self.server.error(f"Chip path {chip_path} does not exist")
        try:
            gpio = periphery.GPIO(
                str(chip_path),
                pin_params["pin_id"],
                pin_params["direction"],
                edge=pin_params.get("edge", "none"),
                bias=pin_params.get("bias", "default"),
                inverted=pin_params["inverted"],
                label="moonraker"
            )
        except Exception:
            logging.exception(
                f"Unable to init {full_name}.  Make sure the gpio is not in "
                "use by another program or exported by sysfs.")
            raise
        return gpio

    def _parse_pin(
        self, pin_desc: str, initial_value: int = 0, req_type: str = "out"
    ) -> Dict[str, Any]:
        params: Dict[str, Any] = {
            "orig": pin_desc,
            "inverted": False,
            "request_type": req_type,
            "initial_value": initial_value
        }
        pin_match = re.match(GPIO_PATTERN, pin_desc, re.VERBOSE)
        if pin_match is None:
            raise self.server.error(
                f"Invalid pin format {pin_desc}. Refer to the configuration "
                "documentation for details on the pin format."
            )
        bias_flag: Optional[str] = pin_match.group("bias")
        params["inverted"] = pin_match.group("inverted") is not None
        if req_type == "event":
            params["direction"] = "in"
            params["edge"] = "both"
            params["bias"] = BIAS_FLAG_TO_DESC[bias_flag or "*"]
        elif req_type == "out":
            if bias_flag is not None:
                raise self.server.error(
                    f"Invalid pin format {pin_desc}.  Bias flag {bias_flag} "
                    "not available for output pins."
                )
            initial_state = bool(initial_value) ^ params["inverted"]
            params["direction"] = "low" if not initial_state else "high"
        chip_id: str = pin_match.group("chip_id") or "gpiochip0"
        pin_name: str = pin_match.group("pin_name")
        params["pin_id"] = int(pin_match.group("pin_id"))
        params["chip_id"] = chip_id
        params["full_name"] = f"{chip_id}:{pin_name}"
        return params

    def close(self) -> None:
        for gpio in self.reserved_gpios.values():
            gpio.close()

class GpioBase:
    def __init__(
        self, gpio: periphery.GPIO, pin_params: Dict[str, Any]
    ) -> None:
        self.orig: str = pin_params["orig"]
        self.name: str = pin_params["full_name"]
        self.inverted: bool = pin_params["inverted"]
        self.gpio = gpio
        self.value: int = pin_params.get("initial_value", 0)

    def close(self) -> None:
        self.gpio.close()

    def is_inverted(self) -> bool:
        return self.inverted

    def get_value(self) -> int:
        return self.value

    def get_name(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.orig

class GpioOutputPin(GpioBase):
    def write(self, value: int) -> None:
        self.value = int(not not value)
        self.gpio.write(bool(self.value))


MAX_ERRORS = 50
ERROR_RESET_TIME = 5.

class GpioEvent(GpioBase):
    def __init__(
        self,
        event_loop: EventLoop,
        gpio: periphery.GPIO,
        pin_params: Dict[str, Any],
        callback: GpioEventCallback
    ) -> None:
        super().__init__(gpio, pin_params)
        self.event_loop = event_loop
        self.callback = callback
        self.on_error: Optional[Callable[[str], None]] = None
        self.debounce_period: float = 0
        self.last_event_time: float = 0.
        self.error_count = 0
        self.last_error_reset = 0.
        self.started = False
        self.debounce_task: Optional[asyncio.Task] = None
        os.set_blocking(self.gpio.fd, False)

    def fileno(self) -> int:
        return self.gpio.fd

    def setup_debounce(
        self, debounce_period: float, err_callback: Optional[Callable[[str], None]]
    ) -> None:
        self.debounce_period = max(debounce_period, 0)
        self.on_error = err_callback

    def start(self) -> None:
        if not self.started:
            self.value = int(self.gpio.read())
            self.last_event_time = self.event_loop.get_loop_time()
            self.event_loop.add_reader(self.gpio.fd, self._on_event_trigger)
            self.started = True
            logging.debug(f"GPIO {self.name}: Listening for events, "
                          f"current state: {self.value}")

    def stop(self) -> None:
        if self.debounce_task is not None:
            self.debounce_task.cancel()
            self.debounce_task = None
        if self.started:
            self.event_loop.remove_reader(self.gpio.fd)
            self.started = False

    def close(self) -> None:
        self.stop()
        self.gpio.close()

    def _on_event_trigger(self) -> None:
        evt = self.gpio.read_event()
        last_value = self.value
        if evt.edge == "rising":     # type: ignore
            self.value = 1
        elif evt.edge == "falling":  # type: ignore
            self.value = 0
        else:
            return
        if self.debounce_period:
            if self.debounce_task is None:
                coro = self._debounce(last_value)
                self.debounce_task = self.event_loop.create_task(coro)
            else:
                self._increment_error()
        elif last_value != self.value:
            # No debounce period and change detected
            self._run_callback()

    async def _debounce(self, last_value: int) -> None:
        await asyncio.sleep(self.debounce_period)
        self.debounce_task = None
        if last_value != self.value:
            self._run_callback()

    def _run_callback(self) -> None:
        eventtime = self.event_loop.get_loop_time()
        evt_duration = eventtime - self.last_event_time
        self.last_event_time = eventtime
        ret = self.callback(eventtime, evt_duration, self.value)
        if ret is not None:
            self.event_loop.create_task(ret)  # type: ignore

    def _increment_error(self) -> None:
        eventtime = self.event_loop.get_loop_time()
        if eventtime - self.last_error_reset > ERROR_RESET_TIME:
            self.error_count = 0
            self.last_error_reset = eventtime
        self.error_count += 1
        if self.error_count >= MAX_ERRORS:
            self.stop()
            if self.on_error is not None:
                self.on_error(
                    f"Too Many Consecutive Errors, GPIO Event Disabled on {self.name}"
                )


def load_component(config: ConfigHelper) -> GpioFactory:
    return GpioFactory(config)



#! /usr/bin/python3
#
# The original enum-converter.py may be found at:
# https://github.com/PackageKit/PackageKit/blob/b64ee9dfa707d5dd2b93c8eebe9930a55fcde108/lib/python/enum-convertor.py
#
# Copyright (C) 2008 - 2012 PackageKit Authors
#
# Licensed under the GNU General Public License Version 2.0
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# The following modifications have been made to the original
# script:
#  * Print the time of conversion
#  * Extract and print the original license of the source
#  * Enumerations are extracted from the header file to preserve
#    order
#  * Use Python "Flag" Enumerations
#  * Extract comments and include them as docstrings
#  * Introduce a string constant validation mode.  This extracts
#    strings from pk-enum.c, then compares to the calculated
#    strings from pk-enum.h.
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# Usage:
#     pk-enum-converter.py pk_enum.h > enums.py
#
# Enum String Validation Mode:
#     pk-enum-converter.py pk_enum.h pk_enum.c
#
# The pk_enum source files, pk-enum.c and pk-enum.h, can be found in the
# PackageKit GitHub repo:
# https://github.com/PackageKit/PackageKit/blob/main/lib/packagekit-glib2/pk-enum.c
# https://github.com/PackageKit/PackageKit/blob/main/lib/packagekit-glib2/pk-enum.h
#
from __future__ import print_function

from re import compile, DOTALL, MULTILINE
import time
import sys
import pathlib
import textwrap

HEADER = \
'''
# This file was autogenerated from %s by pk-enum-converter.py
# on %s UTC
#
# License for original source:
#
%s

from __future__ import annotations
import sys
from enum import Flag, auto

class PkFlag(Flag):
    @classmethod
    def from_pkstring(cls, pkstring: str):
        for name, member in cls.__members__.items():
            if member.pkstring == pkstring:
                return cls(member.value)
        # Return "unknown" flag
        return cls(1)

    @classmethod
    def from_index(cls, index: int):
        return cls(1 << index)

    @property
    def pkstring(self) -> str:
        if self.name is None:
            return " | ".join([f.pkstring for f in self])
        return self.name.lower().replace("_", "-")

    @property
    def desc(self) -> str:
        if self.name is None:
            return ", ".join([f.desc for f in self])
        description = self.name.lower().replace("_", " ")
        return description.capitalize()

    @property
    def index(self) -> int:
        return self.value.bit_length() - 1

    if sys.version_info < (3, 11):
        def __iter__(self):
            for i in range(self._value_.bit_length()):
                val = 1 << i
                if val & self._value_ == val:
                    yield self.__class__(val)
'''  # noqa: E122

FILTER_PKSTRING = \
'''    @property
    def pkstring(self) -> str:
        pks = self.name
        if pks is None:
            return " | ".join([f.pkstring for f in self])
        if pks in ["DEVELOPMENT", "NOT_DEVELOPMENT"]:
            pks = pks[:-6]
        if pks[:4] == "NOT_":
            pks = "~" + pks[4:]
        return pks.lower().replace("_", "-")
'''  # noqa: E122

ERROR_PROPS = \
'''    @property
    def pkstring(self) -> str:
        if self == Error.UPDATE_FAILED_DUE_TO_RUNNING_PROCESS:
            return "failed-due-to-running-process"
        return super().pkstring
'''  # noqa: E122

ALIASES = {
    "Error.OOM": "OUT_OF_MEMORY"
}

header_enum = compile(r"/\*\*\n(.+?)@PK_[A-Z_]+_ENUM_LAST:\s+\*\s+(.+?)"
                      r"\s+\*\*/\s+typedef enum {(.+?)} Pk(.+?)Enum",
                      DOTALL | MULTILINE)
header_value = compile(r"(PK_[A-Z_]+_ENUM)_([A-Z0-9_]+)")
header_desc = compile(r"@PK_[A-Z_]+_ENUM_([A-Z_]+):(.*)")
license = compile(r"(Copyright.+?)\*/", DOTALL | MULTILINE)
enum_h_name = sys.argv[1]
header = pathlib.Path(enum_h_name).read_text()

# Get License
lic_match = license.search(header)
assert lic_match is not None
lic_parts = lic_match.group(1).split("\n")
lic = "\n".join([("# " + p.lstrip("* ")).rstrip(" ") for p in lic_parts])


if len(sys.argv) > 2:
    # Validation Mode, extract strings from the source file, compare to
    # those calculated from the enums in the header file
    enum_to_string = {}
    enum = compile(r"static const PkEnumMatch enum_([^\]]+)\[\] = {(.*?)};",
                   DOTALL | MULTILINE)
    value = compile(r"(PK_[A-Z_]+_ENUM_[A-Z0-9_]+),\s+\"([^\"]+)\"")
    enum_c_name = sys.argv[2]
    inp = pathlib.Path(enum_c_name).read_text()
    for (name, data) in enum.findall(inp):
        for (enum_name, string) in value.findall(data):
            enum_to_string[enum_name] = string
    for (desc_data, comments, data, name) in header_enum.findall(header):
        for (prefix, short_name) in header_value.findall(data):
            if short_name == "LAST":
                continue
            # Validation Mode
            enum_name = f"{prefix}_{short_name}"
            string = enum_to_string[enum_name]
            calc_string = short_name.lower().replace("_", "-")
            if calc_string[:4] == "not-" and name == "Filter":
                calc_string = "~" + calc_string[4:]
            if calc_string != string:
                print(
                    f"Calculated String Mismatch: {name}.{short_name}\n"
                    f"Calculated: {calc_string}\n"
                    f"Extracted: {string}\n")
    exit(0)

print(HEADER % (enum_h_name, time.asctime(time.gmtime()), lic))
# Use the header file for correct enum ordering
for (desc_data, comments, data, name) in header_enum.findall(header):

    print(f"\nclass {name}(PkFlag):")
    # Print Docstring
    print('    """')
    comments = [(" " * 4 + c.lstrip("* ")).rstrip(" ")
                for c in comments.splitlines()]
    for comment in comments:
        comment = comment.expandtabs(4)
        if len(comment) > 79:
            comment = "\n".join(textwrap.wrap(
                comment, 79, subsequent_indent="    ",
                tabsize=4))
        print(comment)
    print("")
    for (item, desc) in header_desc.findall(desc_data):
        line = f"    * {name}.{item}: {desc}".rstrip()
        if len(line) > 79:
            print(f"    * {name}.{item}:")
            print(f"        {desc}")
        else:
            print(line)
    print('    """')
    if name == "Filter":
        print(FILTER_PKSTRING)
    elif name == "Error":
        print(ERROR_PROPS)
    aliases = []
    for (prefix, short_name) in header_value.findall(data):
        if short_name == "LAST":
            continue
        long_name = f"{name}.{short_name}"
        if long_name in ALIASES:
            alias = ALIASES[long_name]
            aliases.append((short_name, alias))
            short_name = alias
        # Print Enums
        print(f"    {short_name} = auto()")
    for name, alias in aliases:
        print(f"    {name} = {alias}")



# Moonraker/Klipper update configuration
#
# Copyright (C) 2022  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import sys
import copy
import pathlib
from ...common import ExtendedEnum
from ...utils import source_info
from typing import (
    TYPE_CHECKING,
    Dict,
    Union,
    List
)

if TYPE_CHECKING:
    from ...confighelper import ConfigHelper
    from ..klippy_connection import KlippyConnection

BASE_CONFIG: Dict[str, Dict[str, str]] = {
    "moonraker": {
        "origin": "https://github.com/arksine/moonraker.git",
        "requirements": "scripts/moonraker-requirements.txt",
        "venv_args": "-p python3",
        "system_dependencies": "scripts/system-dependencies.json",
        "virtualenv": sys.exec_prefix,
        "pip_environment_variables": "SKIP_CYTHON=Y",
        "path": str(source_info.source_path()),
        "managed_services": "moonraker"
    },
    "klipper": {
        "moved_origin": "https://github.com/kevinoconnor/klipper.git",
        "origin": "https://github.com/Klipper3d/klipper.git",
        "requirements": "scripts/klippy-requirements.txt",
        "venv_args": "-p python3",
        "install_script": "scripts/install-octopi.sh",
        "managed_services": "klipper"
    }
}

OPTION_OVERRIDES = ("channel", "pinned_commit", "refresh_interval")

class AppType(ExtendedEnum):
    NONE = 1
    WEB = 2
    GIT_REPO = 3
    ZIP = 4
    PYTHON = 5

    @classmethod
    def detect(cls, app_path: Union[str, pathlib.Path, None] = None):
        # If app path is None, detect Moonraker
        if isinstance(app_path, str):
            app_path = pathlib.Path(app_path).expanduser()
        if source_info.is_git_repo(app_path):
            return AppType.GIT_REPO
        elif app_path is None and source_info.is_vitualenv_project():
            return AppType.PYTHON
        else:
            return AppType.NONE

    @classmethod
    def valid_types(cls) -> List[AppType]:
        all_types = list(cls)
        all_types.remove(AppType.NONE)
        return all_types

    @property
    def supported_channels(self) -> List[Channel]:
        if self == AppType.NONE:
            return []
        elif self in [AppType.WEB, AppType.ZIP]:
            return [Channel.STABLE, Channel.BETA]  # type: ignore
        else:
            return list(Channel)

    @property
    def default_channel(self) -> Channel:
        if self == AppType.GIT_REPO:
            return Channel.DEV  # type: ignore
        return Channel.STABLE  # type: ignore

class Channel(ExtendedEnum):
    STABLE = 1
    BETA = 2
    DEV = 3

def get_base_configuration(config: ConfigHelper) -> ConfigHelper:
    server = config.get_server()
    base_cfg = copy.deepcopy(BASE_CONFIG)
    kconn: KlippyConnection = server.lookup_component("klippy_connection")
    base_cfg["moonraker"]["type"] = str(AppType.detect())
    base_cfg["klipper"]["path"] = str(kconn.path)
    base_cfg["klipper"]["env"] = str(kconn.executable)
    base_cfg["klipper"]["type"] = str(AppType.detect(kconn.path))
    default_channel = config.get("channel", None)
    # Check for configuration overrides
    for app_name in base_cfg.keys():
        if default_channel is not None:
            base_cfg[app_name]["channel"] = default_channel
        override_section = f"update_manager {app_name}"
        if not config.has_section(override_section):
            continue
        app_cfg = config[override_section]
        for opt in OPTION_OVERRIDES:
            if app_cfg.has_option(opt):
                base_cfg[app_name][opt] = app_cfg.get(opt)
    return config.read_supplemental_dict(base_cfg)



from __future__ import annotations
import pytest
import os
import subprocess
import time
import pathlib
import shlex

from typing import Dict, Optional

class KlippyProcess:
    def __init__(self,
                 base_cmd: str,
                 path_args: Dict[str, pathlib.Path],
                 ) -> None:
        self.base_cmd = base_cmd
        self.config_path = path_args['printer.cfg']
        self.orig_config = self.config_path
        self.dict_path = path_args["klipper.dict"]
        self.pty_path = path_args["klippy_pty_path"]
        self.uds_path = path_args["klippy_uds_path"]
        self.proc: Optional[subprocess.Popen] = None
        self.fd: int = -1

    def start(self):
        if self.proc is not None:
            return
        args = (
            f"{self.config_path} -o /dev/null -d {self.dict_path} "
            f"-a {self.uds_path} -I {self.pty_path}"
        )
        cmd = f"{self.base_cmd} {args}"
        cmd_parts = shlex.split(cmd)
        self.proc = subprocess.Popen(cmd_parts)
        for _ in range(250):
            if self.pty_path.exists():
                try:
                    self.fd = os.open(
                        str(self.pty_path), os.O_RDWR | os.O_NONBLOCK)
                except Exception:
                    pass
                else:
                    break
            time.sleep(.01)
        else:
            self.stop()
            pytest.fail("Unable to start Klippy process")
            return False
        return True

    def send_gcode(self, gcode: str) -> None:
        if self.fd == -1:
            return
        try:
            os.write(self.fd, f"{gcode}\n".encode())
        except Exception:
            pass

    def restart(self):
        self.stop()
        self.start()

    def stop(self):
        if self.fd != -1:
            os.close(self.fd)
            self.fd = -1
        if self.proc is not None:
            self.proc.terminate()
            try:
                self.proc.wait(2.)
            except subprocess.TimeoutExpired:
                self.proc.kill()
            self.proc = None

    def get_paths(self) -> Dict[str, pathlib.Path]:
        return {
            "printer.cfg": self.config_path,
            "klipper.dict": self.dict_path,
            "klippy_uds_path": self.uds_path,
            "klippy_pty_path": self.pty_path,
        }



# Support for Moonraker/Klipper/Client announcements
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import datetime
import pathlib
import asyncio
import logging
import email.utils
import xml.etree.ElementTree as etree
from ..common import RequestType
from typing import (
    TYPE_CHECKING,
    Awaitable,
    List,
    Dict,
    Any,
    Optional
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .http_client import HttpClient
    from .database import MoonrakerDatabase


MOONLIGHT_URL = "https://arksine.github.io/moonlight"
UPDATE_CHECK_TIME = 1800.
etree.register_namespace("moonlight", MOONLIGHT_URL)

class Announcements:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.entry_mgr = EntryManager(config)
        self.eventloop = self.server.get_event_loop()
        self.update_timer = self.eventloop.register_timer(
            self._handle_update_timer
        )
        self.request_lock = asyncio.Lock()
        self.dev_mode = config.getboolean("dev_mode", False)
        self.subscriptions: Dict[str, RssFeed] = {
            "moonraker": RssFeed("moonraker", self.entry_mgr, self.dev_mode),
            "klipper": RssFeed("klipper", self.entry_mgr, self.dev_mode)
        }
        self.stored_feeds: List[str] = []
        sub_list: List[str] = config.getlist("subscriptions", [])
        self.configured_feeds: List[str] = ["moonraker", "klipper"]
        for sub in sub_list:
            sub = sub.lower()
            if sub in self.subscriptions:
                continue
            self.configured_feeds.append(sub)
            self.subscriptions[sub] = RssFeed(
                sub, self.entry_mgr, self.dev_mode
            )

        self.server.register_endpoint(
            "/server/announcements/list", RequestType.GET,
            self._list_announcements
        )
        self.server.register_endpoint(
            "/server/announcements/dismiss", RequestType.POST,
            self._handle_dismiss_request
        )
        self.server.register_endpoint(
            "/server/announcements/update", RequestType.POST,
            self._handle_update_request
        )
        self.server.register_endpoint(
            "/server/announcements/feed", RequestType.POST | RequestType.DELETE,
            self._handle_feed_request
        )
        self.server.register_endpoint(
            "/server/announcements/feeds", RequestType.GET,
            self._handle_list_feeds
        )
        self.server.register_notification(
            "announcements:dismissed", "announcement_dismissed"
        )
        self.server.register_notification(
            "announcements:entries_updated", "announcement_update"
        )
        self.server.register_notification(
            "announcements:dismiss_wake", "announcement_wake"
        )

    async def component_init(self) -> None:
        db: MoonrakerDatabase = self.server.lookup_component("database")
        stored_feeds: List[str] = await db.get_item(
            "moonraker", "announcements.stored_feeds", []
        )
        self.stored_feeds = stored_feeds
        for name in stored_feeds:
            if name in self.subscriptions:
                continue
            feed = RssFeed(name, self.entry_mgr, self.dev_mode)
            self.subscriptions[name] = feed
        async with self.request_lock:
            await self.entry_mgr.initialize()
            for sub in self.subscriptions.values():
                await sub.initialize()
        self.update_timer.start()

    async def _handle_update_timer(self, eventtime: float) -> float:
        changed = False
        entries: List[Dict[str, Any]] = []
        async with self.request_lock:
            for sub in self.subscriptions.values():
                ret = await sub.update_entries()
                changed |= ret
            if changed:
                entries = await self.entry_mgr.list_entries()
                self.server.send_event(
                    "announcements:entries_updated", {"entries": entries}
                )
        return eventtime + UPDATE_CHECK_TIME

    async def _handle_dismiss_request(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        async with self.request_lock:
            entry_id: str = web_request.get_str("entry_id")
            wake_time: Optional[int] = web_request.get_int("wake_time", None)
            await self.entry_mgr.dismiss_entry(entry_id, wake_time)
            return {
                "entry_id": entry_id
            }

    async def _list_announcements(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        async with self.request_lock:
            incl_dsm = web_request.get_boolean("include_dismissed", True)
            entries = await self.entry_mgr.list_entries(incl_dsm)
            return {
                "entries": entries,
                "feeds": list(self.subscriptions.keys())
            }

    async def _handle_update_request(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        subs = web_request.get_list("subscriptions", list(self.subscriptions.keys()))
        for sub in subs:
            if sub not in self.subscriptions:
                raise self.server.error(f"No subscription for {sub}")
        async with self.request_lock:
            changed = False
            for sub in subs:
                ret = await self.subscriptions[sub].update_entries()
                changed |= ret
            entries = await self.entry_mgr.list_entries()
            if changed:
                self.eventloop.delay_callback(
                    .05, self.server.send_event,
                    "announcements:entries_updated",
                    {"entries": entries})
            return {
                "entries": entries,
                "modified": changed
            }

    async def _handle_list_feeds(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        return {"feeds": list(self.subscriptions.keys())}

    async def _handle_feed_request(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        req_type = web_request.get_request_type()
        name: str = web_request.get("name")
        name = name.lower()
        changed: bool = False
        db: MoonrakerDatabase = self.server.lookup_component("database")
        result = "skipped"
        if req_type == RequestType.POST:
            if name not in self.subscriptions:
                feed = RssFeed(name, self.entry_mgr, self.dev_mode)
                self.subscriptions[name] = feed
                await feed.initialize()
                changed = await feed.update_entries()
                self.stored_feeds.append(name)
                db.insert_item(
                    "moonraker", "announcements.stored_feeds", self.stored_feeds
                )
                result = "added"
        elif req_type == RequestType.DELETE:
            if name not in self.stored_feeds:
                raise self.server.error(f"Feed '{name}' not stored")
            if name in self.configured_feeds:
                raise self.server.error(
                    f"Feed '{name}' exists in the configuration, cannot remove"
                )
            self.stored_feeds.remove(name)
            db.insert_item(
                "moonraker", "announcements.stored_feeds", self.stored_feeds
            )
            if name in self.subscriptions:
                del self.subscriptions[name]
                changed = await self.entry_mgr.prune_by_feed(name)
                logging.info(f"Removed Announcement Feed: {name}")
                result = "removed"
            else:
                raise self.server.error(f"Feed does not exist: {name}")
        if changed:
            entries = await self.entry_mgr.list_entries()
            self.eventloop.delay_callback(
                .05, self.server.send_event, "announcements:entries_updated",
                {"entries": entries}
            )
        return {
            "feed": name,
            "action": result
        }

    def add_internal_announcement(
        self, title: str, desc: str, url: str, priority: str, feed: str
    ) -> Dict[str, Any]:
        date = datetime.datetime.utcnow()
        entry_id: str = f"{feed}/{date.isoformat(timespec='seconds')}"
        entry = {
            "entry_id": entry_id,
            "url": url,
            "title": title,
            "description": desc,
            "priority": priority,
            "date": date.timestamp(),
            "dismissed": False,
            "date_dismissed": None,
            "dismiss_wake": None,
            "source": "internal",
            "feed": feed
        }
        self.entry_mgr.add_entry(entry)
        self.eventloop.create_task(self._notify_internal())
        return entry

    async def _notify_internal(self) -> None:
        entries = await self.entry_mgr.list_entries()
        self.server.send_event(
            "announcements:entries_updated", {"entries": entries}
        )

    async def remove_announcement(self, entry_id: str) -> None:
        ret = await self.entry_mgr.remove_entry(entry_id)
        if ret is not None:
            entries = await self.entry_mgr.list_entries()
            self.server.send_event(
                "announcements:entries_updated", {"entries": entries}
            )
    async def dismiss_announcement(
        self, entry_id, wake_time: Optional[int] = None
    ) -> None:
        await self.entry_mgr.dismiss_entry(entry_id, wake_time)

    async def get_announcements(
        self, include_dismissed: bool = False
    ) -> List[Dict[str, Any]]:
        return await self.entry_mgr.list_entries(include_dismissed)

    def register_feed(self, name: str) -> None:
        name = name.lower()
        if name in self.subscriptions:
            logging.info(f"Feed {name} already configured")
            return
        logging.info(f"Registering feed {name}")
        self.configured_feeds.append(name)
        self.subscriptions[name] = RssFeed(name, self.entry_mgr, self.dev_mode)

    def close(self):
        self.entry_mgr.close()

class EntryManager:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        database: MoonrakerDatabase
        database = self.server.lookup_component("database")
        database.register_local_namespace("announcements")
        self.announce_db = database.wrap_namespace("announcements")
        self.entry_id_map: Dict[str, str] = {}
        self.next_key = 0
        self.dismiss_handles: Dict[str, asyncio.TimerHandle] = {}

    async def initialize(self) -> None:
        last_key = ""
        eventloop = self.server.get_event_loop()
        curtime = datetime.datetime.utcnow().timestamp()
        for key, entry in await self.announce_db.items():
            last_key = key
            aid = entry["entry_id"]
            self.entry_id_map[aid] = key
            if entry["dismissed"]:
                wake_time: Optional[float] = entry.get("dismiss_wake")
                if wake_time is not None:
                    time_diff = wake_time - curtime
                    if time_diff - 10. < 0.:
                        # announcement is near or past wake time
                        entry["dismissed"] = False
                        entry["date_dismissed"] = None
                        entry["dismiss_wake"] = None
                        self.announce_db[key] = entry
                    else:
                        self.dismiss_handles[key] = eventloop.delay_callback(
                            time_diff, self._wake_dismissed, key
                        )
        if last_key:
            self.next_key = int(last_key, 16) + 1

    async def list_entries(
        self, include_dismissed: bool = True
    ) -> List[Dict[str, Any]]:
        vals: List[Dict[str, Any]] = await self.announce_db.values()
        vals = sorted(vals, key=lambda x: x["date"], reverse=True)
        if include_dismissed:
            return vals
        return [val for val in vals if not val["dismissed"]]

    def has_entry(self, entry_id: str) -> bool:
        return entry_id in self.entry_id_map

    def add_entry(self, entry: Dict[str, Any]) -> Awaitable[None]:
        aid = entry["entry_id"]
        key = f"{self.next_key:06X}"
        self.next_key += 1
        self.entry_id_map[aid] = key
        return self.announce_db.insert(key, entry)

    def remove_entry(self, entry_id: str) -> Awaitable[Any]:
        key = self.entry_id_map.pop(entry_id, None)
        if key is None:
            raise self.server.error(f"No key matching entry id: {entry_id}")
        return self.announce_db.pop(key, None)

    async def dismiss_entry(
        self, entry_id: str, wake_time: Optional[int] = None
    ) -> None:
        key = self.entry_id_map.get(entry_id)
        if key is None:
            raise self.server.error(f"No key matching entry id: {entry_id}")
        entry = await self.announce_db[key]
        is_dismissed = entry["dismissed"]
        if is_dismissed:
            return
        entry["dismissed"] = True
        eventloop = self.server.get_event_loop()
        curtime = datetime.datetime.utcnow().timestamp()
        entry["date_dismissed"] = curtime
        if wake_time is not None:
            entry["dismiss_wake"] = curtime + wake_time
            self.dismiss_handles[key] = eventloop.delay_callback(
                wake_time, self._wake_dismissed, key
            )
        self.announce_db[key] = entry
        eventloop.delay_callback(
            .05, self.server.send_event, "announcements:dismissed",
            {"entry_id": entry_id}
        )

    async def _wake_dismissed(self, key: str) -> None:
        self.dismiss_handles.pop(key, None)
        entry = await self.announce_db.get(key, None)
        if entry is None:
            return
        if not entry["dismissed"]:
            return
        entry["dismissed"] = False
        entry["date_dismissed"] = None
        entry["dismiss_wake"] = None
        self.announce_db[key] = entry
        self.server.send_event(
            "announcements:dismiss_wake", {"entry_id": entry["entry_id"]}
        )

    def prune_by_prefix(self, prefix: str, valid_ids: List[str]) -> bool:
        del_keys: List[str] = []
        for entry_id in list(self.entry_id_map.keys()):
            if not entry_id.startswith(prefix) or entry_id in valid_ids:
                continue
            # Entry is no longer valid and should be removed
            key = self.entry_id_map.pop(entry_id, None)
            if key is not None:
                del_keys.append(key)
        if del_keys:
            self.announce_db.delete_batch(del_keys)
            return True
        return False

    async def prune_by_feed(self, feed: str) -> bool:
        entries = await self.list_entries()
        del_keys: List[str] = []
        for entry in entries:
            if entry["feed"].lower() == feed:
                key = self.entry_id_map.pop(entry["entry_id"], None)
                if key is not None:
                    del_keys.append(key)
        if del_keys:
            self.announce_db.delete_batch(del_keys)
            return True
        return False

    def close(self):
        for handle in self.dismiss_handles.values():
            handle.cancel()

class RssFeed:
    def __init__(
        self, name: str, entry_mgr: EntryManager, dev_mode: bool
    ) -> None:
        self.server = entry_mgr.server
        self.name = name
        self.entry_mgr = entry_mgr
        self.client: HttpClient = self.server.lookup_component("http_client")
        self.database: MoonrakerDatabase
        self.database = self.server.lookup_component("database")
        self.xml_file = f"{self.name}.xml"
        self.asset_url = f"{MOONLIGHT_URL}/assets/{self.xml_file}"
        self.last_modified: int = 0
        self.etag: Optional[str] = None
        self.dev_xml_path: Optional[pathlib.Path] = None
        if dev_mode:
            data_path = pathlib.Path(self.server.get_app_arg("data_path"))
            dev_folder = data_path.joinpath("development/announcements")
            dev_folder.mkdir(parents=True, exist_ok=True)
            self.dev_xml_path = dev_folder.joinpath(self.xml_file)
            logging.info(
                f"Announcement Feed {name}: Dev XML path set to {self.dev_xml_path}"
            )

    async def initialize(self) -> None:
        self.etag = await self.database.get_item(
            "moonraker", f"announcements.{self.name}.etag", None
        )

    async def update_entries(self) -> bool:
        if self.dev_xml_path is None:
            xml_data = await self._fetch_moonlight()
        else:
            xml_data = await self._fetch_local_folder()
        if not xml_data:
            return False
        return self._parse_xml(xml_data)

    async def _fetch_moonlight(self) -> str:
        headers = {"Accept": "application/xml"}
        if self.etag is not None:
            headers["If-None-Match"] = self.etag
        resp = await self.client.get(
            self.asset_url, headers, attempts=5,
            retry_pause_time=.5, enable_cache=False,
        )
        if resp.has_error():
            logging.info(
                f"Failed to update subscription '{self.name}': {resp.error}"
            )
            return ""
        if resp.status_code == 304:
            logging.debug(f"Content at {self.xml_file} not modified")
            return ""
        # update etag
        self.etag = resp.etag
        try:
            if self.etag is not None:
                self.database.insert_item(
                    "moonraker", f"announcements.{self.name}.etag", resp.etag
                )
            else:
                self.database.delete_item(
                    "moonraker", f"announcements.{self.name}.etag",
                )
        except self.server.error:
            pass
        return resp.text

    async def _fetch_local_folder(self) -> str:
        if self.dev_xml_path is None:
            return ""
        if not self.dev_xml_path.is_file():
            logging.info(f"No file at path {self.dev_xml_path}")
            return ""
        mtime = self.dev_xml_path.stat().st_mtime_ns
        if mtime <= self.last_modified:
            logging.debug(f"Content at {self.xml_file} not modified")
            return ""
        try:
            eventloop = self.server.get_event_loop()
            xml_data = await eventloop.run_in_thread(
                self.dev_xml_path.read_text)
        except Exception:
            logging.exception(f"Unable read xml file {self.dev_xml_path}")
            return ""
        self.last_modified = mtime
        return xml_data

    def _parse_xml(self, xml_data: str) -> bool:
        root = etree.fromstring(xml_data)
        channel = root.find("channel")
        if channel is None:
            root_str = etree.tostring(root, encoding="unicode")
            logging.debug(f"Feed {self.name}: no channel found\n{root_str}")
            return False
        # extract prefix
        prefix = channel.findtext("title", "").lower()
        if not prefix:
            logging.info(f"Feed {self.name}: No prefix found")
        items = channel.findall("item")
        valid_ids: List[str] = []
        changed: bool = False
        for item in items:
            guid = item.findtext("guid")
            if guid is None:
                item_str = etree.tostring(item, encoding="unicode")
                logging.debug(f"Feed {self.name}: Invalid Item\n{item_str}")
                continue
            if not prefix:
                # fall back to first guid prefix
                prefix = "/".join(guid.split("/")[:2])
            elif not guid.startswith(prefix):
                logging.debug(
                    f"Feed {self.name}: Guid {guid} is not "
                    f"prefixed with {prefix}")
            valid_ids.append(guid)
            if self.entry_mgr.has_entry(guid):
                continue
            try:
                rfc_date = item.findtext("pubDate", "")
                dt = email.utils.parsedate_to_datetime(rfc_date)
            except Exception:
                dt = datetime.datetime.utcnow()
            entry: Dict[str, Any] = {
                "entry_id": guid,
                "url": item.findtext("link"),
                "title": item.findtext("title"),
                "description": item.findtext("description"),
                "priority": item.findtext("category"),
                "date": dt.timestamp(),
                "dismissed": False,
                "date_dismissed": None,
                "dismiss_wake": None,
                "source": "moonlight",
                "feed": self.name
            }
            changed = True
            self.entry_mgr.add_entry(entry)
        logging.debug(f"Feed {self.name}: found entries {valid_ids}")
        if prefix:
            pruned = self.entry_mgr.prune_by_prefix(prefix, valid_ids)
            changed = changed or pruned
        return changed


def load_component(config: ConfigHelper) -> Announcements:
    return Announcements(config)



from __future__ import annotations
import pytest
import asyncio
import pathlib
from typing import TYPE_CHECKING, Dict
from moonraker.server import ServerError
from moonraker.klippy_connection import KlippyRequest
from mocks import MockReader, MockWriter

if TYPE_CHECKING:
    from server import Server
    from conftest import KlippyProcess

@pytest.mark.usefixtures("klippy")
@pytest.mark.asyncio
async def test_klippy_startup(full_server: Server):
    evtloop = full_server.get_event_loop()
    futs = [evtloop.create_future() for _ in range(3)]
    events = {
        "server:klippy_identified": lambda: futs[0].set_result("id"),
        "server:klippy_started": lambda x: futs[1].set_result("started"),
        "server:klippy_ready": lambda: futs[2].set_result("ready")
    }
    for name, func in events.items():
        full_server.register_event_handler(name, func)
    await full_server.start_server()
    ret = await asyncio.wait_for(asyncio.gather(*futs), 4.)
    assert (
        ret == ["id", "started", "ready"] and
        full_server.klippy_connection.is_connected()
    )

@pytest.mark.asyncio
async def test_gcode_response(ready_server: Server,
                              klippy: KlippyProcess):
    evtloop = ready_server.get_event_loop()
    fut = evtloop.create_future()

    def on_gc_resp(resp: str):
        if not fut.done():
            fut.set_result(resp)
    ready_server.register_event_handler("server:gcode_response", on_gc_resp)
    klippy.send_gcode("M118 Moonraker Test")
    await asyncio.wait_for(fut, 1.)
    assert "Moonraker Test" in fut.result()

@pytest.mark.asyncio
async def test_klippy_shutdown(ready_server: Server, klippy: KlippyProcess):
    evtloop = ready_server.get_event_loop()
    fut = evtloop.create_future()

    def on_shutdown():
        if not fut.done():
            fut.set_result("shutdown")
    ready_server.register_event_handler("server:klippy_shutdown", on_shutdown)
    klippy.send_gcode("M112")
    await asyncio.wait_for(fut, 2.)
    assert fut.result() == "shutdown"

@pytest.mark.asyncio
async def test_klippy_reconnect(ready_server: Server, klippy: KlippyProcess):
    evtloop = ready_server.get_event_loop()
    futs = [evtloop.create_future() for _ in range(2)]
    events = {
        "server:klippy_disconnect": lambda: futs[0].set_result("disconnect"),
        "server:klippy_ready": lambda: futs[1].set_result("ready")
    }
    for name, func in events.items():
        ready_server.register_event_handler(name, func)
    klippy.restart()
    ret = await asyncio.wait_for(asyncio.gather(*futs), 6.)
    assert ret == ["disconnect", "ready"]

@pytest.mark.run_paths(klippy_uds="fake_uds")
@pytest.mark.asyncio
async def test_no_klippy_connection_error(full_server: Server):
    await full_server.start_server()
    with pytest.raises(ServerError):
        kapis = full_server.klippy_connection.klippy_apis
        await kapis.run_gcode("M115")

@pytest.mark.asyncio
async def test_status_update(ready_server: Server, klippy: KlippyProcess):
    evtloop = ready_server.get_event_loop()
    fut = evtloop.create_future()

    def on_status_update(data):
        if not fut.done():
            fut.set_result(data)
    ready_server.register_event_handler("server:status_update",
                                        on_status_update)
    kapis = ready_server.klippy_connection.klippy_apis
    await kapis.subscribe_objects({"toolhead": None})
    klippy.send_gcode("G28")
    await asyncio.wait_for(fut, 2.)
    assert isinstance(fut.result(), dict)

@pytest.mark.run_paths(printer_cfg="error_printer.cfg")
@pytest.mark.asyncio
async def test_klippy_error(ready_server: Server):
    kconn = ready_server.klippy_connection
    assert kconn.state == "error"

@pytest.mark.run_paths(printer_cfg="missing_reqs.cfg")
@pytest.mark.asyncio
async def test_missing_reqs(ready_server: Server):
    mreqs = sorted(ready_server.klippy_connection.missing_requirements)
    expected = ["display_status", "pause_resume", "virtual_sdcard"]
    assert mreqs == expected

@pytest.mark.asyncio
async def test_connection_close(full_server: Server):
    await full_server.start_server()
    # Test multiple close attempts, the second to enter
    # should wait and exit
    ret = full_server.klippy_connection.close(True)
    ret2 = full_server.klippy_connection.close(True)
    await asyncio.wait_for(asyncio.gather(ret, ret2), 4.)
    kconn = full_server.klippy_connection
    assert kconn.connection_task.cancelled()

@pytest.mark.asyncio
async def test_init_error(base_server: Server):
    base_server.server_running = True
    kconn = base_server.klippy_connection

    def mock_is_connected():
        return kconn.init_attempts < 3
    kconn.is_connected = mock_is_connected
    ret = await kconn._init_klippy_connection()
    assert ret is False

def test_connect_fail(base_server: Server):
    ret = base_server.klippy_connection.connect()
    assert ret.result() is False

@pytest.mark.asyncio
async def test_wait_connect_fail(base_server: Server):
    ret = await base_server.klippy_connection.wait_connected()
    assert ret is False

@pytest.mark.run_paths(klippy_uds="fake_uds")
@pytest.mark.asyncio
async def test_no_uds(base_server: Server):
    attempts = [1, 2, 3]

    def mock_is_running():
        attempts.pop(0)
        return len(attempts) > 0
    base_server.is_running = mock_is_running
    ret = await base_server.klippy_connection._do_connect()
    assert ret is False

@pytest.mark.run_paths(klippy_uds="fake_uds")
@pytest.mark.asyncio
async def test_no_uds_access(base_server: Server,
                             path_args: Dict[str, pathlib.Path]):
    attempts = [1, 2, 3]
    uds_path = path_args['klippy_uds_path']
    uds_path.write_text("test")
    uds_path.chmod(mode=222)

    def mock_is_running():
        attempts.pop(0)
        return len(attempts) > 0
    base_server.is_running = mock_is_running
    ret = await base_server.klippy_connection._do_connect()
    assert ret is False

@pytest.mark.asyncio
async def test_write_not_connected(base_server: Server):
    req = KlippyRequest("", {})
    kconn = base_server.klippy_connection
    await kconn._write_request(req)
    assert isinstance(req.response, ServerError)

@pytest.mark.asyncio
async def test_write_error(base_server: Server):
    req = KlippyRequest("", {})
    kconn = base_server.klippy_connection
    kconn.writer = MockWriter()
    await kconn._write_request(req)
    assert isinstance(req.response, ServerError)

@pytest.mark.asyncio
async def test_write_cancelled(base_server: Server):
    req = KlippyRequest("", {})
    kconn = base_server.klippy_connection
    kconn.writer = MockWriter(wait_drain=True)
    task = base_server.event_loop.create_task(kconn._write_request(req))
    base_server.event_loop.delay_callback(.01, task.cancel)
    with pytest.raises(asyncio.CancelledError):
        await task

@pytest.mark.asyncio
async def test_read_error(base_server: Server,
                          caplog: pytest.LogCaptureFixture):
    mock_reader = MockReader("raise_error")
    kconn = base_server.klippy_connection
    await kconn._read_stream(mock_reader)
    assert "Klippy Stream Read Error" == caplog.messages[-1]

@pytest.mark.asyncio
async def test_read_cancelled(base_server: Server):
    mock_reader = MockReader("wait")
    kconn = base_server.klippy_connection
    task = base_server.event_loop.create_task(
        kconn._read_stream(mock_reader))
    base_server.event_loop.delay_callback(.01, task.cancel)
    with pytest.raises(asyncio.CancelledError):
        await task

@pytest.mark.asyncio
async def test_read_decode_error(base_server: Server,
                                 caplog: pytest.LogCaptureFixture):
    mock_reader = MockReader()
    kconn = base_server.klippy_connection
    await kconn._read_stream(mock_reader)
    assert "Error processing Klippy Host Response:" in caplog.messages[-1]

def test_process_unknown_method(base_server: Server,
                                caplog: pytest.LogCaptureFixture):
    cmd = {"method": "test_unknown"}
    kconn = base_server.klippy_connection
    kconn._process_command(cmd)
    assert "Unknown method received: test_unknown" == caplog.messages[-1]

def test_process_unknown_request(base_server: Server,
                                 caplog: pytest.LogCaptureFixture):
    cmd = {"id": 4543}
    kconn = base_server.klippy_connection
    kconn._process_command(cmd)
    expected = f"No request matching request ID: 4543, response: {cmd}"
    assert expected == caplog.messages[-1]

def test_process_invalid_request(base_server: Server):
    req = KlippyRequest("", {})
    kconn = base_server.klippy_connection
    kconn.pending_requests[req.id] = req
    cmd = {"id": req.id}
    kconn._process_command(cmd)
    assert isinstance(req.response, ServerError)

# TODO: This can probably go in a class with test apis
@pytest.mark.asyncio
async def test_call_remote_method(base_server: Server,
                                  klippy: KlippyProcess):
    fut = base_server.get_event_loop().create_future()

    def method_test(result):
        fut.set_result(result)
    base_server.register_remote_method("moonraker_test", method_test)
    base_server.load_components()
    await base_server.server_init()
    ret = base_server.klippy_connection.wait_connected()
    await asyncio.wait_for(ret, 4.)
    klippy.send_gcode("TEST_REMOTE_METHOD")
    await fut
    await base_server._stop_server("terminate")
    assert fut.result() == "test"



# Moonraker Process Stat Tracking
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import asyncio
import struct
import fcntl
import time
import re
import os
import pathlib
import logging
from collections import deque
from ..utils import ioctl_macros
from ..common import RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    Deque,
    Any,
    List,
    Tuple,
    Optional,
    Dict,
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .websockets import WebsocketManager
    STAT_CALLBACK = Callable[[int], Optional[Awaitable]]

VC_GEN_CMD_FILE = "/usr/bin/vcgencmd"
VCIO_PATH = "/dev/vcio"
STATM_FILE_PATH = "/proc/self/smaps_rollup"
NET_DEV_PATH = "/proc/net/dev"
TEMPERATURE_PATH = "/sys/class/thermal/thermal_zone0/temp"
CPU_STAT_PATH = "/proc/stat"
MEM_AVAIL_PATH = "/proc/meminfo"
STAT_UPDATE_TIME = 1.
REPORT_QUEUE_SIZE = 30
THROTTLE_CHECK_INTERVAL = 10
WATCHDOG_REFRESH_TIME = 2.
REPORT_BLOCKED_TIME = 4.

THROTTLED_FLAGS = {
    1: "Under-Voltage Detected",
    1 << 1: "Frequency Capped",
    1 << 2: "Currently Throttled",
    1 << 3: "Temperature Limit Active",
    1 << 16: "Previously Under-Volted",
    1 << 17: "Previously Frequency Capped",
    1 << 18: "Previously Throttled",
    1 << 19: "Previously Temperature Limited"
}

class ProcStats:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.event_loop = self.server.get_event_loop()
        self.watchdog = Watchdog(self)
        self.stat_update_timer = self.event_loop.register_timer(
            self._handle_stat_update)
        self.vcgencmd: Optional[VCGenCmd] = None
        if os.path.exists(VC_GEN_CMD_FILE) and os.path.exists(VCIO_PATH):
            logging.info("Detected 'vcgencmd', throttle checking enabled")
            self.vcgencmd = VCGenCmd()
            self.server.register_notification("proc_stats:cpu_throttled")
        else:
            logging.info("Unable to find 'vcgencmd', throttle checking "
                         "disabled")
        self.temp_file = pathlib.Path(TEMPERATURE_PATH)
        self.smaps = pathlib.Path(STATM_FILE_PATH)
        self.netdev_file = pathlib.Path(NET_DEV_PATH)
        self.cpu_stats_file = pathlib.Path(CPU_STAT_PATH)
        self.meminfo_file = pathlib.Path(MEM_AVAIL_PATH)
        self.server.register_endpoint(
            "/machine/proc_stats", RequestType.GET, self._handle_stat_request
        )
        self.server.register_event_handler(
            "server:klippy_shutdown", self._handle_shutdown
        )
        self.server.register_notification("proc_stats:proc_stat_update")
        self.proc_stat_queue: Deque[Dict[str, Any]] = deque(maxlen=30)
        self.last_update_time = time.time()
        self.last_proc_time = time.process_time()
        self.throttle_check_lock = asyncio.Lock()
        self.total_throttled: int = 0
        self.last_throttled: int = 0
        self.update_sequence: int = 0
        self.last_net_stats: Dict[str, Dict[str, Any]] = {}
        self.last_cpu_stats: Dict[str, Tuple[int, int]] = {}
        self.cpu_usage: Dict[str, float] = {}
        self.memory_usage: Dict[str, int] = {}
        self.stat_callbacks: List[STAT_CALLBACK] = []

    async def component_init(self) -> None:
        self.stat_update_timer.start()
        self.watchdog.start()

    def register_stat_callback(self, callback: STAT_CALLBACK) -> None:
        self.stat_callbacks.append(callback)

    async def _handle_stat_request(self,
                                   web_request: WebRequest
                                   ) -> Dict[str, Any]:
        ts: Optional[Dict[str, Any]] = None
        if self.vcgencmd is not None:
            ts = await self._check_throttled_state()
        cpu_temp = await self.event_loop.run_in_thread(
            self._get_cpu_temperature)
        wsm: WebsocketManager = self.server.lookup_component("websockets")
        websocket_count = wsm.get_count()
        return {
            'moonraker_stats': list(self.proc_stat_queue),
            'throttled_state': ts,
            'cpu_temp': cpu_temp,
            'network': self.last_net_stats,
            'system_cpu_usage': self.cpu_usage,
            'system_uptime': time.clock_gettime(time.CLOCK_BOOTTIME),
            'system_memory': self.memory_usage,
            'websocket_connections': websocket_count
        }

    async def _handle_shutdown(self) -> None:
        msg = "\nMoonraker System Usage Statistics:"
        for stats in self.proc_stat_queue:
            msg += f"\n{self._format_stats(stats)}"
        cpu_temp = await self.event_loop.run_in_thread(
            self._get_cpu_temperature)
        msg += f"\nCPU Temperature: {cpu_temp}"
        logging.info(msg)
        if self.vcgencmd is not None:
            ts = await self._check_throttled_state()
            logging.info(f"Throttled Flags: {' '.join(ts['flags'])}")

    async def _handle_stat_update(self, eventtime: float) -> float:
        update_time = eventtime
        proc_time = time.process_time()
        time_diff = update_time - self.last_update_time
        usage = round((proc_time - self.last_proc_time) / time_diff * 100, 2)
        cpu_temp, mem, mem_units, net = (
            await self.event_loop.run_in_thread(self._read_system_files)
        )
        for dev in net:
            bytes_sec = 0.
            if dev in self.last_net_stats:
                last_dev_stats = self.last_net_stats[dev]
                cur_total: int = net[dev]['rx_bytes'] + net[dev]['tx_bytes']
                last_total: int = last_dev_stats['rx_bytes'] + \
                    last_dev_stats['tx_bytes']
                bytes_sec = round((cur_total - last_total) / time_diff, 2)
            net[dev]['bandwidth'] = bytes_sec
        self.last_net_stats = net
        result = {
            'time': time.time(),
            'cpu_usage': usage,
            'memory': mem,
            'mem_units': mem_units
        }
        self.proc_stat_queue.append(result)
        wsm: WebsocketManager = self.server.lookup_component("websockets")
        websocket_count = wsm.get_count()
        self.server.send_event("proc_stats:proc_stat_update", {
            'moonraker_stats': result,
            'cpu_temp': cpu_temp,
            'network': net,
            'system_cpu_usage': self.cpu_usage,
            'system_memory': self.memory_usage,
            'websocket_connections': websocket_count
        })
        if (
            not self.update_sequence % THROTTLE_CHECK_INTERVAL
            and self.vcgencmd is not None
        ):
            ts = await self._check_throttled_state()
            cur_throttled = ts['bits']
            if cur_throttled & ~self.total_throttled:
                self.server.add_log_rollover_item(
                    'throttled', f"CPU Throttled Flags: {ts['flags']}")
            if cur_throttled != self.last_throttled:
                self.server.send_event("proc_stats:cpu_throttled", ts)
            self.last_throttled = cur_throttled
            self.total_throttled |= cur_throttled
        for cb in self.stat_callbacks:
            ret = cb(self.update_sequence)
            if ret is not None:
                await ret
        self.last_update_time = update_time
        self.last_proc_time = proc_time
        self.update_sequence += 1
        return eventtime + STAT_UPDATE_TIME

    async def _check_throttled_state(self) -> Dict[str, Any]:
        ret = {'bits': 0, 'flags': ["?"]}
        if self.vcgencmd is not None:
            async with self.throttle_check_lock:
                try:
                    resp = await self.event_loop.run_in_thread(self.vcgencmd.run)
                    ret["bits"] = tstate = int(resp.strip().split("=")[-1], 16)
                    ret["flags"] = [
                        desc for flag, desc in THROTTLED_FLAGS.items() if flag & tstate
                    ]
                except Exception:
                    pass
        return ret

    def _read_system_files(self) -> Tuple:
        mem, units = self._get_memory_usage()
        temp = self._get_cpu_temperature()
        net_stats = self._get_net_stats()
        self._update_cpu_stats()
        self._update_system_memory()
        return temp, mem, units, net_stats

    def _get_memory_usage(self) -> Tuple[Optional[int], Optional[str]]:
        try:
            mem_data = self.smaps.read_text()
            rss_match = re.search(r"Rss:\s+(\d+)\s+(\w+)", mem_data)
            if rss_match is None:
                return None, None
            mem = int(rss_match.group(1))
            units = rss_match.group(2)
        except Exception:
            return None, None
        return mem, units

    def _get_cpu_temperature(self) -> Optional[float]:
        try:
            res = int(self.temp_file.read_text().strip())
            temp = res / 1000.
        except Exception:
            return None
        return temp

    def _get_net_stats(self) -> Dict[str, Any]:
        net_stats: Dict[str, Any] = {}
        try:
            ret = self.netdev_file.read_text()
            dev_info = re.findall(r"([\w]+):(.+)", ret)
            for (dev_name, stats) in dev_info:
                parsed_stats = stats.strip().split()
                net_stats[dev_name] = {
                    'rx_bytes': int(parsed_stats[0]),
                    'tx_bytes': int(parsed_stats[8]),
                    'rx_packets': int(parsed_stats[1]),
                    'tx_packets': int(parsed_stats[9]),
                    'rx_errs': int(parsed_stats[2]),
                    'tx_errs': int(parsed_stats[10]),
                    'rx_drop': int(parsed_stats[3]),
                    'tx_drop': int(parsed_stats[11])
                }
            return net_stats
        except Exception:
            return {}

    def _update_system_memory(self) -> None:
        mem_stats: Dict[str, Any] = {}
        try:
            ret = self.meminfo_file.read_text()
            total_match = re.search(r"MemTotal:\s+(\d+)", ret)
            avail_match = re.search(r"MemAvailable:\s+(\d+)", ret)
            if total_match is not None and avail_match is not None:
                mem_stats["total"] = int(total_match.group(1))
                mem_stats["available"] = int(avail_match.group(1))
                mem_stats["used"] = mem_stats["total"] - mem_stats["available"]
            self.memory_usage.update(mem_stats)
        except Exception:
            pass

    def _update_cpu_stats(self) -> None:
        try:
            cpu_usage: Dict[str, Any] = {}
            ret = self.cpu_stats_file.read_text()
            usage_info: List[str] = re.findall(r"cpu[^\n]+", ret)
            for cpu in usage_info:
                parts = cpu.split()
                name = parts[0]
                cpu_sum = sum([int(t) for t in parts[1:]])
                cpu_idle = int(parts[4])
                if name in self.last_cpu_stats:
                    last_sum, last_idle = self.last_cpu_stats[name]
                    cpu_delta = cpu_sum - last_sum
                    idle_delta = cpu_idle - last_idle
                    cpu_used = cpu_delta - idle_delta
                    cpu_usage[name] = round(
                        100 * (cpu_used / cpu_delta), 2)
                self.cpu_usage = cpu_usage
                self.last_cpu_stats[name] = (cpu_sum, cpu_idle)
        except Exception:
            pass

    def _format_stats(self, stats: Dict[str, Any]) -> str:
        return f"System Time: {stats['time']:2f}, " \
               f"Usage: {stats['cpu_usage']}%, " \
               f"Memory: {stats['memory']} {stats['mem_units']}"

    def log_last_stats(self, count: int = 1):
        count = min(len(self.proc_stat_queue), count)
        msg = ""
        for stats in list(self.proc_stat_queue)[-count:]:
            msg += f"\n{self._format_stats(stats)}"
        logging.info(msg)

    def close(self) -> None:
        self.stat_update_timer.stop()
        self.watchdog.stop()

class Watchdog:
    def __init__(self, proc_stats: ProcStats) -> None:
        self.proc_stats = proc_stats
        self.event_loop = proc_stats.event_loop
        self.blocked_count: int = 0
        self.last_watch_time: float = 0.
        self.watchdog_timer = self.event_loop.register_timer(
            self._watchdog_callback
        )

    def _watchdog_callback(self, eventtime: float) -> float:
        time_diff = eventtime - self.last_watch_time
        if time_diff > REPORT_BLOCKED_TIME:
            self.blocked_count += 1
            logging.info(
                f"EVENT LOOP BLOCKED: {round(time_diff, 2)} seconds"
                f", total blocked count: {self.blocked_count}")
            # delay the stat logging so we capture the CPU percentage after
            # the next cycle
            self.event_loop.delay_callback(
                .2, self.proc_stats.log_last_stats, 5)
        self.last_watch_time = eventtime
        return eventtime + WATCHDOG_REFRESH_TIME

    def start(self):
        if not self.watchdog_timer.is_running():
            self.last_watch_time = self.event_loop.get_loop_time()
            self.watchdog_timer.start()

    def stop(self):
        self.watchdog_timer.stop()

class VCGenCmd:
    """
    This class uses the BCM2835 Mailbox to directly query the throttled
    state.  This should be less resource intensive than calling "vcgencmd"
    in a subprocess.
    """
    MAX_STRING_SIZE = 1024
    GET_RESULT_CMD = 0x00030080
    UINT_SIZE = struct.calcsize("@I")
    def __init__(self) -> None:
        self.cmd_struct = struct.Struct(f"@6I{self.MAX_STRING_SIZE}sI")
        self.cmd_buf = bytearray(self.cmd_struct.size)
        self.mailbox_req = ioctl_macros.IOWR(100, 0, "c_char_p")
        self.err_logged: bool = False

    def run(self, cmd: str = "get_throttled") -> str:
        try:
            fd = os.open(VCIO_PATH, os.O_RDWR)
            self.cmd_struct.pack_into(
                self.cmd_buf, 0,
                self.cmd_struct.size,
                0x00000000,
                self.GET_RESULT_CMD,
                self.MAX_STRING_SIZE,
                0,
                0,
                cmd.encode("utf-8"),
                0x00000000
            )
            fcntl.ioctl(fd, self.mailbox_req, self.cmd_buf)
        except OSError:
            if not self.err_logged:
                logging.exception("VCIO vcgencmd failed")
                self.err_logged = True
            return ""
        finally:
            os.close(fd)
        result = self.cmd_struct.unpack_from(self.cmd_buf)
        ret: int = result[5]
        if ret:
            logging.info(f"vcgencmd returned {ret}")
        resp: bytes = result[6]
        null_index = resp.find(b'\x00')
        if null_index <= 0:
            return ""
        return resp[:null_index].decode()

def load_component(config: ConfigHelper) -> ProcStats:
    return ProcStats(config)



from __future__ import annotations
import pytest
import json
import asyncio
import tornado.websocket

from typing import (
    TYPE_CHECKING,
    Union,
    Tuple,
    Callable,
    Dict,
    List,
    Any,
    Optional,
)

if TYPE_CHECKING:
    from tornado.websocket import WebSocketClientConnection

class WebsocketError(Exception):
    def __init__(self, code, *args: object) -> None:
        super().__init__(*args)
        self.code = code

class WebsocketClient:
    error = WebsocketError
    def __init__(self,
                 type: str = "ws",
                 port: int = 7010
                 ) -> None:
        self.ws: Optional[WebSocketClientConnection] = None
        self.pending_requests: Dict[int, asyncio.Future] = {}
        self.notify_cbs: Dict[str, List[Callable[..., None]]] = {}
        assert type in ["ws", "wss"]
        self.url = f"{type}://127.0.0.1:{port}/websocket"

    async def connect(self, token: Optional[str] = None) -> None:
        url = self.url
        if token is not None:
            url += f"?token={token}"
        self.ws = await tornado.websocket.websocket_connect(
            url, connect_timeout=2.,
            on_message_callback=self._on_message_received)

    async def request(self,
                      remote_method: str,
                      args: Dict[str, Any] = {}
                      ) -> Dict[str, Any]:
        if self.ws is None:
            pytest.fail("Websocket Not Connected")
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        req, req_id = self._encode_request(remote_method, args)
        self.pending_requests[req_id] = fut
        await self.ws.write_message(req)
        return await asyncio.wait_for(fut, 2.)

    def _encode_request(self,
                        method: str,
                        args: Dict[str, Any]
                        ) -> Tuple[str, int]:
        request: Dict[str, Any] = {
            'jsonrpc': "2.0",
            'method': method,
        }
        if args:
            request['params'] = args
        req_id = id(request)
        request["id"] = req_id
        return json.dumps(request), req_id

    def _on_message_received(self, message: Union[str, bytes, None]) -> None:
        if isinstance(message, str):
            self._decode_jsonrpc(message)

    def _decode_jsonrpc(self, data: str) -> None:
        try:
            resp: Dict[str, Any] = json.loads(data)
        except json.JSONDecodeError:
            pytest.fail(f"Websocket JSON Decode Error: {data}")
        header = resp.get('jsonrpc', "")
        if header != "2.0":
            # Invalid Json, set error if we can get the id
            pytest.fail(f"Invalid jsonrpc header: {data}")
        req_id: Optional[int] = resp.get("id")
        method: Optional[str] = resp.get("method")
        if method is not None:
            if req_id is None:
                params = resp.get("params", [])
                if not isinstance(params, list):
                    pytest.fail("jsonrpc notification params"
                                f"should always be a list: {data}")
                if method in self.notify_cbs:
                    for func in self.notify_cbs[method]:
                        func(*params)
            else:
                # This is a request from the server (should not happen)
                pytest.fail(f"Server should not request from client: {data}")
        elif req_id is not None:
            pending_fut = self.pending_requests.pop(req_id, None)
            if pending_fut is None:
                # No future pending for this response
                return
            # This is a response
            if "result" in resp:
                pending_fut.set_result(resp["result"])
            elif "error" in resp:
                err = resp["error"]
                try:
                    code = err["code"]
                    msg = err["message"]
                except Exception:
                    pytest.fail(f"Invalid jsonrpc error: {data}")
                exc = WebsocketError(code, msg)
                pending_fut.set_exception(exc)
            else:
                pytest.fail(
                    f"Invalid jsonrpc packet, no result or error: {data}")
        else:
            # Invalid json
            pytest.fail(f"Invalid jsonrpc packet, no id: {data}")

    def register_notify_callback(self, name: str, callback) -> None:
        if name in self.notify_cbs:
            self.notify_cbs[name].append(callback)
        else:
            self.notify_cbs[name][callback]

    def close(self):
        for fut in self.pending_requests.values():
            if not fut.done():
                fut.set_exception(WebsocketError(
                    0, "Closing Websocket Client"))
        if self.ws is not None:
            self.ws.close(1000, "Test Complete")



# PanelDue LCD display support
#
# Copyright (C) 2020  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import serial
import os
import time
import errno
import logging
import asyncio
from collections import deque
from ..utils import ServerError
from ..utils import json_wrapper as jsonw

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Deque,
    Any,
    Tuple,
    Optional,
    Dict,
    List,
    Callable,
    Coroutine,
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from .klippy_connection import KlippyConnection
    from .klippy_apis import KlippyAPI as APIComp
    from .file_manager.file_manager import FileManager as FMComp
    FlexCallback = Callable[..., Optional[Coroutine]]

MIN_EST_TIME = 10.
INITIALIZE_TIMEOUT = 10.

class PanelDueError(ServerError):
    pass


RESTART_GCODES = ["RESTART", "FIRMWARE_RESTART"]

class SerialConnection:
    def __init__(self,
                 config: ConfigHelper,
                 paneldue: PanelDue
                 ) -> None:
        self.event_loop = config.get_server().get_event_loop()
        self.paneldue = paneldue
        self.port: str = config.get('serial')
        self.baud = config.getint('baud', 57600)
        self.partial_input: bytes = b""
        self.ser: Optional[serial.Serial] = None
        self.fd: Optional[int] = None
        self.connected: bool = False
        self.send_busy: bool = False
        self.send_buffer: bytes = b""
        self.attempting_connect: bool = True

    def disconnect(self, reconnect: bool = False) -> None:
        if self.connected:
            if self.fd is not None:
                self.event_loop.remove_reader(self.fd)
                self.fd = None
            self.connected = False
            if self.ser is not None:
                self.ser.close()
            self.ser = None
            self.partial_input = b""
            self.send_buffer = b""
            self.paneldue.initialized = False
            logging.info("PanelDue Disconnected")
        if reconnect and not self.attempting_connect:
            self.attempting_connect = True
            self.event_loop.delay_callback(1., self.connect)

    async def connect(self) -> None:
        self.attempting_connect = True
        start_time = connect_time = time.time()
        while not self.connected:
            if connect_time > start_time + 30.:
                logging.info("Unable to connect, aborting")
                break
            logging.info(f"Attempting to connect to: {self.port}")
            try:
                # XXX - sometimes the port cannot be exclusively locked, this
                # would likely be due to a restart where the serial port was
                # not correctly closed.  Maybe don't use exclusive mode?
                self.ser = serial.Serial(
                    self.port, self.baud, timeout=0, exclusive=True)
            except (OSError, IOError, serial.SerialException):
                logging.exception(f"Unable to open port: {self.port}")
                await asyncio.sleep(2.)
                connect_time += time.time()
                continue
            self.fd = self.ser.fileno()
            fd = self.fd = self.ser.fileno()
            os.set_blocking(fd, False)
            self.event_loop.add_reader(fd, self._handle_incoming)
            self.connected = True
            logging.info("PanelDue Connected")
        self.attempting_connect = False

    def _handle_incoming(self) -> None:
        # Process incoming data using same method as gcode.py
        if self.fd is None:
            return
        try:
            data = os.read(self.fd, 4096)
        except os.error:
            return

        if not data:
            # possibly an error, disconnect
            self.disconnect(reconnect=True)
            logging.info("serial_display: No data received, disconnecting")
            return

        # Remove null bytes, separate into lines
        data = data.strip(b'\x00')
        lines = data.split(b'\n')
        lines[0] = self.partial_input + lines[0]
        self.partial_input = lines.pop()
        for line in lines:
            try:
                decoded_line = line.strip().decode('utf-8', 'ignore')
                self.paneldue.process_line(decoded_line)
            except ServerError:
                logging.exception(
                    f"GCode Processing Error: {decoded_line}")
                self.paneldue.handle_gcode_response(
                    f"!! GCode Processing Error: {decoded_line}")
            except Exception:
                logging.exception("Error during gcode processing")

    def send(self, data: bytes) -> None:
        self.send_buffer += data
        if not self.send_busy:
            self.send_busy = True
            self.event_loop.register_callback(self._do_send)

    async def _do_send(self) -> None:
        assert self.fd is not None
        while self.send_buffer:
            if not self.connected:
                break
            try:
                sent = os.write(self.fd, self.send_buffer)
            except os.error as e:
                if e.errno == errno.EBADF or e.errno == errno.EPIPE:
                    sent = 0
                else:
                    await asyncio.sleep(.001)
                    continue
            if sent:
                self.send_buffer = self.send_buffer[sent:]
            else:
                logging.exception(
                    "Error writing data, closing serial connection")
                self.disconnect(reconnect=True)
                return
        self.send_busy = False

class PanelDue:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.event_loop = self.server.get_event_loop()
        self.file_manager: FMComp = self.server.lookup_component('file_manager')
        self.klippy_apis: APIComp = self.server.lookup_component('klippy_apis')
        self.kinematics: str = "none"
        self.machine_name = config.get('machine_name', "Klipper")
        self.firmware_name: str = "Repetier | Klipper"
        self.last_message: Optional[str] = None
        self.last_gcode_response: Optional[str] = None
        self.current_file: str = ""
        self.file_metadata: Dict[str, Any] = {}
        self.enable_checksum = config.getboolean('enable_checksum', True)
        self.debug_queue: Deque[str] = deque(maxlen=100)

        # Initialize tracked state.
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        self.printer_state: Dict[str, Dict[str, Any]] = kconn.get_subscription_cache()
        self.extruder_count: int = 0
        self.heaters: List[str] = []
        self.is_ready: bool = False
        self.is_shutdown: bool = False
        self.initialized: bool = False
        self.cq_busy: bool = False
        self.gq_busy: bool = False
        self.command_queue: List[Tuple[FlexCallback, Any, Any]] = []
        self.gc_queue: List[str] = []
        self.last_printer_state: str = 'O'
        self.last_update_time: float = 0.

        # Set up macros
        self.confirmed_gcode: str = ""
        self.mbox_sequence: int = 0
        self.available_macros: Dict[str, str] = {}
        self.confirmed_macros = {
            "RESTART": "RESTART",
            "FIRMWARE_RESTART": "FIRMWARE_RESTART"}
        macros = config.getlist('macros', None)
        if macros is not None:
            # The macro's configuration name is the key, whereas the full
            # command is the value
            self.available_macros = {m.split()[0]: m for m in macros}
        conf_macros = config.getlist('confirmed_macros', None)
        if conf_macros is not None:
            # The macro's configuration name is the key, whereas the full
            # command is the value
            self.confirmed_macros = {m.split()[0]: m for m in conf_macros}
        self.available_macros.update(self.confirmed_macros)
        self.non_trivial_keys = config.getlist('non_trivial_keys', ["Klipper state"])
        self.ser_conn = SerialConnection(config, self)
        logging.info("PanelDue Configured")

        # Register server events
        self.server.register_event_handler(
            "server:klippy_ready", self._process_klippy_ready
        )
        self.server.register_event_handler(
            "server:klippy_shutdown", self._process_klippy_shutdown
        )
        self.server.register_event_handler(
            "server:klippy_disconnect", self._process_klippy_disconnect
        )
        self.server.register_event_handler(
            "server:gcode_response", self.handle_gcode_response
        )
        self.server.register_remote_method("paneldue_beep", self.paneldue_beep)

        # These commands are directly executued on the server and do not to
        # make a request to Klippy
        self.direct_gcodes: Dict[str, FlexCallback] = {
            'M20': self._run_paneldue_M20,
            'M30': self._run_paneldue_M30,
            'M36': self._run_paneldue_M36,
            'M408': self._run_paneldue_M408
        }

        # These gcodes require special parsing or handling prior to being
        # sent via Klippy's "gcode/script" api command.
        self.special_gcodes: Dict[str, Callable[[List[str]], str]] = {
            'M0': lambda args: "CANCEL_PRINT",
            'M23': self._prepare_M23,
            'M24': lambda args: "RESUME",
            'M25': lambda args: "PAUSE",
            'M32': self._prepare_M32,
            'M98': self._prepare_M98,
            'M120': lambda args: "SAVE_GCODE_STATE STATE=PANELDUE",
            'M121': lambda args: "RESTORE_GCODE_STATE STATE=PANELDUE",
            'M290': self._prepare_M290,
            'M292': self._prepare_M292,
            'M999': lambda args: "FIRMWARE_RESTART"
        }

    async def component_init(self) -> None:
        await self.ser_conn.connect()

    async def _process_klippy_ready(self) -> None:
        # Request "info" and "configfile" status
        retries = 10
        printer_info: Dict[str, Any] = {}
        cfg_status: Dict[str, Any] = {}
        while retries:
            try:
                printer_info = await self.klippy_apis.get_klippy_info()
                cfg_status = await self.klippy_apis.query_objects({'configfile': None})
            except self.server.error:
                logging.exception("PanelDue initialization request failed")
                retries -= 1
                if not retries:
                    raise
                await asyncio.sleep(1.)
                continue
            break

        self.firmware_name = "Repetier | Klipper " + printer_info['software_version']
        config: Dict[str, Any] = cfg_status.get('configfile', {}).get('config', {})
        printer_cfg: Dict[str, Any] = config.get('printer', {})
        self.kinematics = printer_cfg.get('kinematics', "none")

        logging.info(
            f"PanelDue Config Received:\n"
            f"Firmware Name: {self.firmware_name}\n"
            f"Kinematics: {self.kinematics}\n"
            f"Printer Config: {config}\n")

        # Make subscription request
        sub_args: Dict[str, Optional[List[str]]] = {
            "motion_report": None,
            "gcode_move": None,
            "toolhead": None,
            "virtual_sdcard": None,
            "fan": None,
            "display_status": None,
            "print_stats": None,
            "idle_timeout": None,
            "gcode_macro PANELDUE_BEEP": None
        }
        self.extruder_count = 0
        self.heaters = []
        extruders = []
        for cfg in config:
            if cfg.startswith("extruder"):
                self.extruder_count += 1
                extruders.append(cfg)
                sub_args[cfg] = None
            elif cfg == "heater_bed":
                self.heaters.append(cfg)
                sub_args[cfg] = None
        extruders.sort()
        self.heaters.extend(extruders)
        try:
            await self.klippy_apis.subscribe_objects(sub_args)
        except self.server.error:
            logging.exception("Unable to complete subscription request")
        self.is_shutdown = False
        self.is_ready = True

    def _process_klippy_shutdown(self) -> None:
        self.is_shutdown = True

    def _process_klippy_disconnect(self) -> None:
        # Tell the PD that the printer is "off"
        self.write_response({'status': 'O'})
        self.last_printer_state = 'O'
        self.is_ready = False
        self.is_shutdown = self.is_shutdown = False

    def paneldue_beep(self, frequency: int, duration: float) -> None:
        duration = int(duration * 1000.)
        self.write_response(
            {'beep_freq': frequency, 'beep_length': duration})

    def process_line(self, line: str) -> None:
        self.debug_queue.append(line)
        # If we find M112 in the line then skip verification
        if "M112" in line.upper():
            self.event_loop.register_callback(self.klippy_apis.emergency_stop)
            return

        if self.enable_checksum:
            # Get line number
            line_index = line.find(' ')
            try:
                line_no: Optional[int] = int(line[1:line_index])
            except Exception:
                line_index = -1
                line_no = None

            # Verify checksum
            cs_index = line.rfind('*')
            try:
                checksum = int(line[cs_index+1:])
            except Exception:
                # Invalid checksum, do not process
                msg = "!! Invalid Checksum"
                if line_no is not None:
                    msg += f" Line Number: {line_no}"
                logging.exception("PanelDue: " + msg)
                raise PanelDueError(msg)

            # Checksum is calculated by XORing every byte in the line other
            # than the checksum itself
            calculated_cs = 0
            for c in line[:cs_index]:
                calculated_cs ^= ord(c)
            if calculated_cs & 0xFF != checksum:
                msg = "!! Invalid Checksum"
                if line_no is not None:
                    msg += f" Line Number: {line_no}"
                logging.info("PanelDue: " + msg)
                raise PanelDueError(msg)

            script = line[line_index+1:cs_index]
        else:
            script = line
        # Execute the gcode.  Check for special RRF gcodes that
        # require special handling
        parts = script.split()
        cmd = parts[0].strip()
        if cmd in ["M23", "M30", "M32", "M36", "M37", "M98"]:
            arg = script[len(cmd):].strip()
            parts = [cmd, arg]

        # Check for commands that query state and require immediate response
        if cmd in self.direct_gcodes:
            params: Dict[str, Any] = {}
            for p in parts[1:]:
                if p[0] not in "PSR":
                    params["arg_p"] = p.strip(" \"\t\n")
                    continue
                arg = p[0].lower()
                try:
                    val = int(p[1:].strip()) if arg in "sr" \
                        else p[1:].strip(" \"\t\n")
                except Exception:
                    msg = f"paneldue: Error parsing direct gcode {script}"
                    self.handle_gcode_response("!! " + msg)
                    logging.exception(msg)
                    return
                params[f"arg_{arg}"] = val
            func = self.direct_gcodes[cmd]
            self.queue_command(func, **params)
            return

        # Prepare GCodes that require special handling
        if cmd in self.special_gcodes:
            sgc_func = self.special_gcodes[cmd]
            script = sgc_func(parts[1:])

        if not script:
            return
        self.queue_gcode(script)

    def queue_gcode(self, script: str) -> None:
        self.gc_queue.append(script)
        if not self.gq_busy:
            self.gq_busy = True
            self.event_loop.register_callback(self._process_gcode_queue)

    async def _process_gcode_queue(self) -> None:
        while self.gc_queue:
            script = self.gc_queue.pop(0)
            try:
                if script in RESTART_GCODES:
                    await self.klippy_apis.do_restart(script)
                else:
                    await self.klippy_apis.run_gcode(script)
            except self.server.error:
                msg = f"Error executing script {script}"
                self.handle_gcode_response("!! " + msg)
                logging.exception(msg)
        self.gq_busy = False

    def queue_command(self, cmd: FlexCallback, *args, **kwargs) -> None:
        self.command_queue.append((cmd, args, kwargs))
        if not self.cq_busy:
            self.cq_busy = True
            self.event_loop.register_callback(self._process_command_queue)

    async def _process_command_queue(self) -> None:
        while self.command_queue:
            cmd, args, kwargs = self.command_queue.pop(0)
            try:
                ret = cmd(*args, **kwargs)
                if ret is not None:
                    await ret
            except Exception:
                logging.exception("Error processing command")
        self.cq_busy = False

    def _clean_filename(self, filename: str) -> str:
        # Remove quotes and whitespace
        filename.strip(" \"\t\n")
        # Remove drive number
        if filename.startswith("0:/"):
            filename = filename[3:]
        # Remove initial "gcodes" folder.  This is necessary
        # due to the HACK in the paneldue_M20 gcode.
        if filename.startswith("gcodes/"):
            filename = filename[6:]
        elif filename.startswith("/gcodes/"):
            filename = filename[7:]
        # Start with a "/" so the gcode parser can correctly
        # handle files that begin with digits or special chars
        if filename[0] != "/":
            filename = "/" + filename
        return filename

    def _prepare_M23(self, args: List[str]) -> str:
        filename = self._clean_filename(args[0])
        return f"M23 {filename}"

    def _prepare_M32(self, args: List[str]) -> str:
        filename = self._clean_filename(args[0])
        # Escape existing double quotes in the file name
        filename = filename.replace("\"", "\\\"")
        return f"SDCARD_PRINT_FILE FILENAME=\"{filename}\""

    def _prepare_M98(self, args: List[str]) -> str:
        macro = args[0][1:].strip(" \"\t\n")
        name_start = macro.rfind('/') + 1
        macro = macro[name_start:]
        cmd = self.available_macros.get(macro)
        if cmd is None:
            raise PanelDueError(f"Macro {macro} invalid")
        if macro in self.confirmed_macros:
            self._create_confirmation(macro, cmd)
            cmd = ""
        return cmd

    def _prepare_M290(self, args: List[str]) -> str:
        # args should in in the format Z0.02
        offset = args[0][1:].strip()
        return f"SET_GCODE_OFFSET Z_ADJUST={offset} MOVE=1"

    def _prepare_M292(self, args: List[str]) -> str:
        p_val = int(args[0][1])
        if p_val == 0:
            cmd = self.confirmed_gcode
            self.confirmed_gcode = ""
            return cmd
        return ""

    def _create_confirmation(self, name: str, gcode: str) -> None:
        self.mbox_sequence += 1
        self.confirmed_gcode = gcode
        title = "Confirmation Dialog"
        msg = f"Please confirm your intent to run {name}."  \
            " Press OK to continue, or CANCEL to abort."
        mbox: Dict[str, Any] = {}
        mbox['msgBox.mode'] = 3
        mbox['msgBox.msg'] = msg
        mbox['msgBox.seq'] = self.mbox_sequence
        mbox['msgBox.title'] = title
        mbox['msgBox.controls'] = 0
        mbox['msgBox.timeout'] = 0
        logging.debug(f"Creating PanelDue Confirmation: {mbox}")
        self.write_response(mbox)

    def handle_gcode_response(self, response: str) -> None:
        # Only queue up "non-trivial" gcode responses.  At the
        # moment we'll handle state changes and errors
        if "Klipper state" in response \
                or response.startswith('!!'):
            self.last_gcode_response = response
        else:
            for key in self.non_trivial_keys:
                if key in response:
                    self.last_gcode_response = response
                    return

    def write_response(self, response: Dict[str, Any]) -> None:
        byte_resp = jsonw.dumps(response) + b"\r\n"
        self.ser_conn.send(byte_resp)

    def _get_printer_status(self) -> str:
        # PanelDue States applicable to Klipper:
        # I = idle, P = printing from SD, S = stopped (shutdown),
        # C = starting up (not ready), A = paused, D = pausing,
        # B = busy
        if self.is_shutdown:
            return 'S'

        p_state = self.printer_state
        sd_state: str
        sd_state = p_state.get("print_stats", {}).get("state", "standby")
        if sd_state == "printing":
            if self.last_printer_state == 'A':
                # Resuming
                return 'R'
            # Printing
            return 'P'
        elif sd_state == "paused":
            p_active = (
                p_state.get("idle_timeout", {}).get("state", 'Idle') == "Printing"
            )
            if p_active and self.last_printer_state != 'A':
                # Pausing
                return 'D'
            else:
                # Paused
                return 'A'

        return 'I'

    def _run_paneldue_M408(self,
                           arg_r: Optional[int] = None,
                           arg_s: int = 1
                           ) -> None:
        response: Dict[str, Any] = {}
        sequence = arg_r
        response_type = arg_s

        curtime = self.event_loop.get_loop_time()
        if curtime - self.last_update_time > INITIALIZE_TIMEOUT:
            self.initialized = False
        self.last_update_time = curtime

        if not self.initialized:
            response['dir'] = "/macros"
            response['files'] = list(self.available_macros.keys())
            self.initialized = True
        if not self.is_ready:
            self.last_printer_state = 'O'
            response['status'] = self.last_printer_state
            self.write_response(response)
            return
        if sequence is not None and self.last_gcode_response:
            # Send gcode responses
            response['seq'] = sequence + 1
            response['resp'] = self.last_gcode_response
            self.last_gcode_response = None
        if response_type == 1:
            # Extended response Request
            response['myName'] = self.machine_name
            response['firmwareName'] = self.firmware_name
            response['numTools'] = self.extruder_count
            response['geometry'] = self.kinematics
            response['axes'] = 3

        p_state = self.printer_state
        toolhead = p_state.get("toolhead", {})
        gcode_move = p_state.get("gcode_move", {})
        self.last_printer_state = self._get_printer_status()
        response['status'] = self.last_printer_state
        response['babystep'] = round(
            gcode_move.get('homing_origin', [0., 0., 0., 0.])[2], 3
        )

        # Current position
        pos: List[float]
        homed_pos: str
        sfactor: float
        pos = p_state.get("motion_report", {}).get('live_position', [0., 0., 0., 0.])
        response['pos'] = [round(p, 2) for p in pos[:3]]
        homed_pos = toolhead.get('homed_axes', "")
        response['homed'] = [int(a in homed_pos) for a in "xyz"]
        sfactor = round(gcode_move.get('speed_factor', 1.) * 100, 2)
        response['sfactor'] = sfactor

        # Print Progress Tracking
        sd_status = p_state.get('virtual_sdcard', {})
        print_stats = p_state.get('print_stats', {})
        fname: str = print_stats.get('filename', "")
        sd_print_state: Optional[str] = print_stats.get('state')
        if sd_print_state in ['printing', 'paused']:
            # We know a file has been loaded, initialize metadata
            if self.current_file != fname:
                self.current_file = fname
                self.file_metadata = self.file_manager.get_file_metadata(fname)
            progress: float = sd_status.get('progress', 0)
            # progress and print tracking
            if progress:
                response['fraction_printed'] = round(progress, 3)
                est_time: float = self.file_metadata.get('estimated_time', 0)
                if est_time > MIN_EST_TIME:
                    # file read estimate
                    times_left = [int(est_time - est_time * progress)]
                    # filament estimate
                    est_total_fil: Optional[float]
                    est_total_fil = self.file_metadata.get('filament_total')
                    if est_total_fil:
                        cur_filament: float = print_stats.get(
                            'filament_used', 0.)
                        fpct = min(1., cur_filament / est_total_fil)
                        times_left.append(int(est_time - est_time * fpct))
                    # object height estimate
                    obj_height: Optional[float]
                    obj_height = self.file_metadata.get('object_height')
                    if obj_height:
                        cur_height: float = gcode_move.get(
                            'gcode_position', [0., 0., 0., 0.]
                        )[2]
                        hpct = min(1., cur_height / obj_height)
                        times_left.append(int(est_time - est_time * hpct))
                else:
                    # The estimated time is not in the metadata, however we
                    # can still provide an estimate based on file progress
                    duration: float = print_stats.get('print_duration', 0.)
                    times_left = [int(duration / progress - duration)]
                response['timesLeft'] = times_left
        else:
            # clear filename and metadata
            self.current_file = ""
            self.file_metadata = {}

        fan_speed: Optional[float] = p_state.get('fan', {}).get('speed')
        if fan_speed is not None:
            response['fanPercent'] = [round(fan_speed * 100, 1)]

        extruder_name: str = ""
        if self.extruder_count > 0:
            extruder_name = toolhead.get('extruder', "")
            if extruder_name:
                tool = 0
                if extruder_name != "extruder":
                    tool = int(extruder_name[-1])
                response['tool'] = tool

        # Report Heater Status
        efactor: float = round(gcode_move.get('extrude_factor', 1.) * 100., 2)

        for name in self.heaters:
            htr_state = p_state.get(name, {})
            temp: float = round(htr_state.get('temperature', 0.0), 1)
            target: float = round(htr_state.get('target', 0.0), 1)
            response.setdefault('heaters', []).append(temp)
            response.setdefault('active', []).append(target)
            response.setdefault('standby', []).append(target)
            if name.startswith('extruder'):
                a_stat = 2 if name == extruder_name else 1
                response.setdefault('hstat', []).append(a_stat if target else 0)
                response.setdefault('efactor', []).append(efactor)
                response.setdefault('extr', []).append(round(pos[3], 2))
            else:
                response.setdefault('hstat', []).append(2 if target else 0)

        # Display message (via M117)
        msg: str = p_state.get('display_status', {}).get('message', "")
        if msg and msg != self.last_message:
            response['message'] = msg
            # reset the message so it only shows once.  The paneldue
            # is strange about this, and displays it as a full screen
            # notification
        self.last_message = msg
        self.write_response(response)

    def _run_paneldue_M20(self, arg_p: str, arg_s: int = 0) -> None:
        response_type = arg_s
        if response_type != 2:
            logging.info(
                f"Cannot process response type {response_type} in M20")
            return
        path = arg_p

        # Strip quotes if they exist
        path = path.strip('\"')

        # Path should come in as "0:/macros, or 0:/<gcode_folder>".  With
        # repetier compatibility enabled, the default folder is root,
        # ie. "0:/"
        if path.startswith("0:/"):
            path = path[2:]
        response: Dict[str, Any] = {'dir': path}
        response['files'] = []

        if path == "/macros":
            response['files'] = list(self.available_macros.keys())
        else:
            # HACK: The PanelDue has a bug where it does not correctly detect
            # subdirectories if we return the root as "/".  Moonraker can
            # support a "gcodes" directory, however we must choose between this
            # support or disabling RRF specific gcodes (this is done by
            # identifying as Repetier).
            # The workaround below converts both "/" and "/gcodes" paths to
            # "gcodes".
            if path == "/":
                response['dir'] = "/gcodes"
                path = "gcodes"
            elif path.startswith("/gcodes"):
                path = path[1:]

            flist = self.file_manager.list_dir(path, simple_format=True)
            if flist:
                response['files'] = flist
        self.write_response(response)

    async def _run_paneldue_M30(self, arg_p: str = "") -> None:
        # Delete a file.  Clean up the file name and make sure
        # it is relative to the "gcodes" root.
        path = arg_p
        path = path.strip('\"')
        if path.startswith("0:/"):
            path = path[3:]
        elif path[0] == "/":
            path = path[1:]

        if not path.startswith("gcodes/"):
            path = "gcodes/" + path
        await self.file_manager.delete_file(path)

    def _run_paneldue_M36(self, arg_p: Optional[str] = None) -> None:
        response: Dict[str, Any] = {}
        filename: Optional[str] = arg_p
        sd_status = self.printer_state.get('virtual_sdcard', {})
        print_stats = self.printer_state.get('print_stats', {})
        if filename is None:
            # PanelDue is requesting file information on a
            # currently printed file
            active = False
            if sd_status and print_stats:
                filename = print_stats['filename']
                active = sd_status['is_active']
            if not filename or not active:
                # Either no file printing or no virtual_sdcard
                response['err'] = 1
                self.write_response(response)
                return
            else:
                response['fileName'] = filename.split("/")[-1]

        # For consistency make sure that the filename begins with the
        # "gcodes/" root.  The M20 HACK should add this in some cases.
        # Ideally we would add support to the PanelDue firmware that
        # indicates Moonraker supports a "gcodes" directory.
        if filename[0] == "/":
            filename = filename[1:]
        if not filename.startswith("gcodes/"):
            filename = "gcodes/" + filename

        metadata: Dict[str, Any] = \
            self.file_manager.get_file_metadata(filename)
        if metadata:
            response['err'] = 0
            response['size'] = metadata['size']
            # workaround for PanelDue replacing the first "T" found
            response['lastModified'] = "T" + time.ctime(metadata['modified'])
            slicer: Optional[str] = metadata.get('slicer')
            if slicer is not None:
                response['generatedBy'] = slicer
            height: Optional[float] = metadata.get('object_height')
            if height is not None:
                response['height'] = round(height, 2)
            layer_height: Optional[float] = metadata.get('layer_height')
            if layer_height is not None:
                response['layerHeight'] = round(layer_height, 2)
            filament: Optional[float] = metadata.get('filament_total')
            if filament is not None:
                response['filament'] = [round(filament, 1)]
            est_time: Optional[float] = metadata.get('estimated_time')
            if est_time is not None:
                response['printTime'] = int(est_time + .5)
        else:
            response['err'] = 1
        self.write_response(response)

    def close(self) -> None:
        self.ser_conn.disconnect()
        msg = "\nPanelDue GCode Dump:"
        for i, gc in enumerate(self.debug_queue):
            msg += f"\nSequence {i}: {gc}"
        logging.debug(msg)

def load_component(config: ConfigHelper) -> PanelDue:
    return PanelDue(config)
